/*=============================================

	PC-6031 Emulator Firmware
	
	2013.2.9 by Koichi Nishida
	Copyright (C) 2013 Koichi NISHIDA
	email : tulip-house@msf.biglobe.ne.jp

=============================================*/

/*
hardware information of SD6031
use ATMEGA164P AVR.
connect 20MHz crystal to the AVR.

the fuse setting :
(note that JTAG must be disable by this setting)

 LOW:  11011110
 HIGH: 11011001

port direction and connection

DDRB = SD card (CLK,DO,DI,CS,EJECT,WP),LED2,LED1
DDRA = 0b11111111; data out to P6
DDRC = 0b00000000; data in from P6
DDRD = 0b11110000; ctrl out to P6 (ATN,DAC,RFD,DAV), ctrl in from P6 (ATN,DAC,RFD,DAV)
*/

#include <avr/io.h>
#include <avr/interrupt.h>
#include <string.h>

#define WAIT 1

#define min(a,b) ((a)>(b)?(b):(a))

// for selecting SD card
#define SD_CS	4	// PB4 SS	CS

// ==================== prototypes ====================

// calculate 2^x
unsigned char pow2(register unsigned char x);

#define DRIVE_NUM 4

// ---------- for SD card and image ----------

// SD card information 
struct SD_info {
	unsigned char inited;
	unsigned long bpbAddr, rootAddr;
	unsigned long fatAddr;				// the beginning of FAT
	unsigned char sectorsPerCluster;	// sectors per cluster
	unsigned short sectorsPerFat;	
	unsigned long userAddr;				// the beginning of user data
	unsigned short drive_start_cluster[DRIVE_NUM];
};

// disk image information
#define FAT_ELEMS 64
struct image_info {
	unsigned short start_cluster[DRIVE_NUM];	// start cluster (0 if not mounted)
	unsigned short fat[DRIVE_NUM][FAT_ELEMS];	// the fat for the image
	unsigned char prevFatNum[DRIVE_NUM];		// previous fat number for the mage
	char protect[DRIVE_NUM];					// write protect
	unsigned long len[DRIVE_NUM];				// image length
};

// write a byte data to the SD card
void SD_writeByte(unsigned char c);

// read data from the SD card
unsigned char SD_readByte(void);

// wait until finish a command
void SD_waitFinish(void);

// issue SD card command without getting response
void SD_cmd_(unsigned char cmd, unsigned long adr);

// issue SD card command and wait normal response
void SD_cmd(unsigned char cmd, unsigned long adr);

// get command response from the SD card
unsigned char getResp(void);

// issue command 17 and get ready for reading
void SD_cmd17(unsigned long adr);

// find the file and create the image information
char SD_findImage(unsigned short dir_cluster, char *name, char *ext, char isDir,
	char *protect, unsigned long *length, unsigned short *start_cluster);

// send P31 image file names to P6
void sendFileNames(unsigned short dir_cluster);
	
// prepare the FAT table on memory
void SD_prepareFatSub(unsigned short start_cluster, unsigned short *fat, unsigned short len,
	unsigned char fatNum, unsigned char fatElemNum);
unsigned char SD_prepareFat(unsigned short *long_sector, unsigned short *ft,
	unsigned char track, unsigned char sector, unsigned char drive);
	
// initialization called from SD_checkEject
void SD_init(void);

// called when the SD card is inserted or removed
void SD_checkEject(void);

// ---------- communication with P6  ----------

// communicate with PC-6001
// send data to PC-6001
void P6_snd(unsigned char d1, unsigned char d2);
// receive data to PC-6001
void P6_rcv(unsigned char *d1, unsigned char *d2);

// ---------- PC-6031 status ----------
struct pc6031_status {
	unsigned char drive;
	unsigned char track;
	unsigned char sector;
	unsigned char length;
	unsigned char valid;
	unsigned char mode;
	unsigned char error;
};

// initialize PC-6031 status
void pc6031_init(void);

// process PC-6031 command
void pc6031_process(void);

//==================== implementation ====================

struct SD_info SD;				// SD card information
struct image_info image;		// image information
struct pc6031_status pc6031;	// PC-6031 status
unsigned char LED;				// LED status
unsigned char RST_init;

// calculate 2^x
unsigned char pow2(register unsigned char x)
{
	register unsigned char i, r=1;

	for (i=0; i!=x; i++) {
		r<<=1;
	}
	return r;
}

// write a byte data to the SD card
void SD_writeByte(unsigned char c)
{
	SPDR = c;
	// Wait for transmission complete
	while (!(SPSR & (1<<SPIF)))
		if (bit_is_set(PINB,3)) return;
}

// read data from the SD card
unsigned char SD_readByte(void)
{
	SPDR = 0xFF;
	// Wait for reception complete
	while (!(SPSR & (1<<SPIF)))
		if (bit_is_set(PINB,3)) return 0xff;
	// Return Data Register
	return SPDR;
}

// wait until data is written to the SD card
void SD_waitFinish(void)
{
	unsigned char ch;

	do {
		ch = SD_readByte();
		if (bit_is_set(PINB,3)) return;
	} while (ch != 0xff);
}

// issue a SD card command without getting response
void SD_cmd_(unsigned char cmd, unsigned long adr)
{
	SD_writeByte(0xff);
	SD_writeByte(0x40+cmd);
	SD_writeByte(adr>>24);
	SD_writeByte((adr>>16)&0xff);
	SD_writeByte((adr>>8)&0xff);
	SD_writeByte(adr&0xff);
	SD_writeByte(0x95);
	SD_writeByte(0xff);
}

// issue a SD card command and wait normal response
void SD_cmd(unsigned char cmd, unsigned long adr)
{
	unsigned char res;

	do {
		SD_cmd_(cmd, adr);
	} while (((res=getResp())!=0) && (res!=0xff));
}

// get a command response from the SD card
unsigned char getResp(void)
{
	unsigned char ch;

	do {
		ch = SD_readByte();
		if (bit_is_set(PINB,3)) return 0xff;
	} while ((ch&0x80) != 0);
	return ch;
}

// issue command 17 and get ready for reading
void SD_cmd17(unsigned long adr)
{
	unsigned char ch;

	SD_cmd(17, adr);
	do {	
		ch = SD_readByte();
		if (bit_is_set(PINB,3)) return;
	} while (ch != 0xfe);
}

// find the file extension and create the image information
char SD_findImage(unsigned short dir_cluster, char *name, char *ext, char isDir, char *protect,
	unsigned long *length, unsigned short *start_cluster)
{
	unsigned char i, j, s;
	unsigned long max_entry_adr = 0;
	unsigned short max_time = 0, max_date = 0;
	char isRoot = !dir_cluster;
	unsigned char spc = isRoot?64:SD.sectorsPerCluster;
	unsigned short ft = dir_cluster;
	unsigned long entry_adr = isRoot?SD.rootAddr:
		(unsigned long)SD.userAddr+((unsigned long)(ft-2)*SD.sectorsPerCluster)*512;
		
	if (bit_is_set(PINB,3)) return 0;

	// find extension
	do {
		for (s=0; s!=spc; s++) {
			for (i=0; i!=8; i++, entry_adr+=32) {
				char ext_[4], name_[9], d;
				unsigned char f_time[2], f_date[2];

				// file name
				SD_cmd(16, 8);
				SD_cmd17(entry_adr);
				for (j=0; j!=8; j++) name_[j] = SD_readByte();
				name_[8] = 0;
				SD_readByte(); SD_readByte(); // discard CRC bytes

				// check first char
				d = name_[0];
				
				if (d==0x00) goto SDF_EXIT;
				if (/*(d==0x05)||*/(d==0x2e)||(d==0xe5)) continue;
				//if (!(((d>='A')&&(d<='Z'))||((d>='0')&&(d<='9')))) continue;

				// check attribute
				SD_cmd(16, 1);
				SD_cmd17(entry_adr+11);
				d = SD_readByte();
				SD_readByte(); SD_readByte(); // discard CRC bytes
				
				if ((d==0xf) || (d==0x8)) continue;
				if (isDir) {if (!(d&0x10)||(d&0x0e)) continue;}
				else {if (d&0x1e) continue;}
		
				// check extension & protect
				SD_cmd(16, 4);
				SD_cmd17(entry_adr+8);
				for (j=0; j!=3; j++) ext_[j] = SD_readByte();
				ext_[3] = 0;
				if (protect) *protect = ((SD_readByte()&1)<<3); else SD_readByte();
				SD_readByte(); SD_readByte(); // discard CRC bytes

				// check time stamp
				SD_cmd(16, 4);
				SD_cmd17(entry_adr+22);
				f_time[0] = SD_readByte(); f_time[1] = SD_readByte();
				f_date[0] = SD_readByte(); f_date[1] = SD_readByte();
				SD_readByte(); SD_readByte(); // discard CRC bytes
				if ((!ext || !strcmp(ext, ext_))&&(!name || !strcmp(name, name_))) {
					unsigned short tm = f_time[0] + f_time[1]*256;
					unsigned short dt = f_date[0] + f_date[1]*256;
					
					if ((dt>max_date)||((dt==max_date)&&(tm>=max_time))) {
						max_time = tm;
						max_date = dt;
						max_entry_adr = entry_adr;
					}
				}			
			}
		}
		if (!isRoot) {
			SD_cmd(16, 2);
			SD_cmd17((unsigned long)SD.fatAddr+(unsigned long)ft*2);
			ft = SD_readByte();
			ft += (unsigned short)SD_readByte()*0x100;
			entry_adr = (unsigned long)SD.userAddr+((unsigned long)(ft-2)*SD.sectorsPerCluster)*512;
			SD_readByte(); SD_readByte(); // discard CRC bytes
		}
	} while (!isRoot);

SDF_EXIT:
	
	if (max_entry_adr) {
		if (length) {
			unsigned char *tmp = (unsigned char *)length;
			// set length
			SD_cmd(16, 4);
			SD_cmd17(max_entry_adr+28);
			for (j=0; j<4; j++) *(tmp++) = SD_readByte();
			SD_readByte(); SD_readByte();
		}
		if (start_cluster) {
			// set start cluster number
			SD_cmd(16, 2);
			SD_cmd17(max_entry_adr+26);
			*start_cluster = SD_readByte();
			*start_cluster += (unsigned short)SD_readByte()*0x100;
			SD_readByte(); SD_readByte(); // discard CRC bytes
		}
		if (protect) {
			SD_cmd(16, 1);
			SD_cmd17(max_entry_adr+11);
			*protect = ((SD_readByte()&1)<<3);
			SD_readByte(); SD_readByte(); // discard CRC bytes
		}
		return 1;
	}
	return 0;
	// if 0 then not found...
}

// send P31 image file names to P6
void sendFileNames(unsigned short dir_cluster)
{
	unsigned char i, j, s;
	unsigned char spc = SD.sectorsPerCluster;
	unsigned short ft = dir_cluster;
	unsigned long entry_adr =
		(unsigned long)SD.userAddr+((unsigned long)(ft-2)*SD.sectorsPerCluster)*512;
		
	if (bit_is_set(PINB,3)) return;

	// find extension
	do {
		for (s=0; s!=spc; s++) {
			for (i=0; i!=8; i++, entry_adr+=32) {
				char ext_[4], name_[9], d;

				// file name
				SD_cmd(16, 8);
				SD_cmd17(entry_adr);
				for (j=0; j!=8; j++) name_[j] = SD_readByte();
				name_[8] = 0;
				SD_readByte(); SD_readByte(); // discard CRC bytes

				// check first char
				d = name_[0];
				
				if (d==0x00) {
					P6_snd(0, 0);
					return;
				}
				if (/*(d==0x05)||*/(d==0x2e)||(d==0xe5)) continue;
				//if (!(((d>='A')&&(d<='Z'))||((d>='0')&&(d<='9')))) continue;

				// check attribute
				SD_cmd(16, 1);
				SD_cmd17(entry_adr+11);
				d = SD_readByte();
				SD_readByte(); SD_readByte(); // discard CRC bytes
				
				if ((d==0xf) || (d==0x8)) continue;
				if (d&0x1e) continue;
		
				// check extension
				SD_cmd(16, 3);
				SD_cmd17(entry_adr+8);
				for (j=0; j!=3; j++) ext_[j] = SD_readByte();
				ext_[3] = 0;
				SD_readByte(); SD_readByte(); // discard CRC bytes

				if (!strcmp("P31", ext_)) {
					P6_snd(1, 0);
					for (j=0; j!=8; j++) {
						P6_snd(name_[j], 0);
					}
				}			
			}
		}
		SD_cmd(16, 2);
		SD_cmd17((unsigned long)SD.fatAddr+(unsigned long)ft*2);
		ft = SD_readByte();
		ft += (unsigned short)SD_readByte()*0x100;
		entry_adr = (unsigned long)SD.userAddr+((unsigned long)(ft-2)*SD.sectorsPerCluster)*512;
		SD_readByte(); SD_readByte(); // discard CRC bytes
	} while (1);
}

// prepare a FAT table on memory
unsigned char SD_prepareFat(unsigned short *long_sector, unsigned short *ft,
	unsigned char track, unsigned char sector, unsigned char drive)
{
	unsigned char fatNum;
	unsigned short long_cluster;
	
	*long_sector = (unsigned short)track*16+(sector-1);
	long_cluster = (*long_sector)/SD.sectorsPerCluster;
	fatNum = long_cluster/FAT_ELEMS;
	if (fatNum != image.prevFatNum[drive]) {
		unsigned short sectLen = (image.len[drive]+511)/512;

		image.prevFatNum[drive] = fatNum;
		SD_prepareFatSub(image.start_cluster[drive], image.fat[drive],
			(sectLen+SD.sectorsPerCluster-1)/SD.sectorsPerCluster, fatNum, FAT_ELEMS);
	}
	*ft = image.fat[drive][long_cluster%FAT_ELEMS];

	return ((*ft == 0xfff7) ? 0 : 1);
}

void SD_prepareFatSub(unsigned short start_cluster, unsigned short *fat, unsigned short len,
	unsigned char fatNum, unsigned char fatElemNum)
{
	unsigned short ft;
	unsigned char fn;
	unsigned short i;

	for (i=0; i<FAT_ELEMS; i++) {
		fat[i] = 0xfff7;		// initialize
	}

	if (bit_is_set(PINB,3)) return;

	ft = start_cluster;
	if (0==fatNum) fat[0] = ft;
	for (i=0; i<len; i++) {
		fn = (i+1)/fatElemNum;
		SD_cmd(16, 2);
		SD_cmd17((unsigned long)SD.fatAddr+(unsigned long)ft*2);
		ft = SD_readByte();
		ft += (unsigned short)SD_readByte()*0x100;
		SD_readByte(); SD_readByte(); // discard CRC bytes
		if (fn==fatNum) fat[(i+1)%fatElemNum] = ft;
		if ((ft>0xfff6)||(fn>fatNum)) break;
	}
	SD_cmd(16, (unsigned long)512);	
}

// initialization called from SD_checkEject
void SD_init(void)
{
	unsigned char ch;
	unsigned short i;
	char str[9];

	//SPI enable, master mode, f/128
	SPCR = ((1<<SPE)|(1<<MSTR)|(1<<SPR1)|(1<<SPR0));

	// initialize the SD card
	PORTB |= (1<<SD_CS);						// disable CS
	for (i = 0; i != 10; i++)
		SD_writeByte(0xFF);						// Send 10 * 8 = 80 clock
	PORTB &= ~(1<<SD_CS);						// enable CS
	for(i=0; i < 2; i++) SD_writeByte(0xFF);	// Send 2 * 8 = 16 clock

	SD_cmd_(0, 0);	// command 0	
 	do {	
		if (bit_is_set(PINB,3)) return;	
		ch = SD_readByte();
	} while (ch != 0x01);
	PORTB |= (1<<SD_CS);
	while (1) {
		if (bit_is_set(PINB,3)) return;
		PORTB &= ~(1<<SD_CS);
		SD_cmd_(55, 0);	// command 55
		ch = getResp();
		if (ch == 0xff) return;
		if (ch & 0xfe) continue;
		// if (ch == 0x00) break;
		PORTB |= (1<<SD_CS);
		PORTB &= ~(1<<SD_CS);
		SD_cmd_(41, 0);	// command 41	
		if (!(ch=getResp())) break;
		if (ch == 0xff) return;
		PORTB |= (1<<SD_CS);
	}

	//SPI enable, master mode, f/2
	SPCR = ((1<<SPE)|(1<<MSTR));
	SPSR = (1<<SPI2X);

	// BPB address
	SD_cmd(16,5);
	SD_cmd17(54);
	for (i=0; i<5; i++) str[i] = SD_readByte();
	SD_readByte(); SD_readByte(); // discard CRC
	if ((str[0]=='F')&&(str[1]=='A')&&(str[2]=='T')&&
		(str[3]=='1')&&(str[4]=='6')) {
		SD.bpbAddr = 0;
	} else {
		SD_cmd(16, 4);
		SD_cmd17((unsigned long)0x1c6);
		SD.bpbAddr = SD_readByte();
		SD.bpbAddr += (unsigned long)SD_readByte()*0x100;
		SD.bpbAddr += (unsigned long)SD_readByte()*0x10000;
		SD.bpbAddr += (unsigned long)SD_readByte()*0x1000000;
		SD.bpbAddr *= 512;
		SD_readByte(); SD_readByte(); // discard CRC bytes
	}
	if (bit_is_set(PINB,3)) return;

	// sectorsPerCluster and reservedSectors
	{
		unsigned short reservedSectors;
		SD_cmd(16, 3);
		SD_cmd17(SD.bpbAddr+0xd);		
		SD.sectorsPerCluster = SD_readByte();
		reservedSectors = SD_readByte();
		reservedSectors += (unsigned short)SD_readByte()*0x100;
		SD_readByte(); SD_readByte(); // discard CRC bytes	
		// sectorsPerCluster = 0x40 at 2GB, 0x10 at 512MB
		// reservedSectors = 2 at 2GB
		SD.fatAddr = SD.bpbAddr + (unsigned long)512*reservedSectors;
	}
	if (bit_is_set(PINB,3)) return;

	{
		// sectorsPerFat and rootAddr
		SD_cmd(16, 2);
		SD_cmd17(SD.bpbAddr+0x16);
		SD.sectorsPerFat = SD_readByte();
		SD.sectorsPerFat += (unsigned short)SD_readByte()*0x100;
		SD_readByte(); SD_readByte(); // discard CRC bytes		
		// sectorsPerFat =  at 512MB,  0xEF at 2GB
		SD.rootAddr = SD.fatAddr + ((unsigned long)SD.sectorsPerFat*2*512);
		SD.userAddr = SD.rootAddr+(unsigned long)512*32;
	}
	if (bit_is_set(PINB,3)) return;

	for (i=0; i!=8; i++) str[i] = ' ';
	str[0]='D'; str[8]=0;
	if (RST_init) pc6031.mode = 0;
	for (i=0; i!=DRIVE_NUM; i++) {
		image.prevFatNum[i] = 0xff;
		image.start_cluster[i] = 0;
		str[1]='0'+(i+1);
		if (SD_findImage(0, str, (char *)0, 1, (char *)0, (unsigned long *)0, &(SD.drive_start_cluster[i]))) {
			if (!RST_init ||  !SD_findImage(SD.drive_start_cluster[i], "INIT    ", "P31", 0, &(image.protect[i]), &(image.len[i]), &(image.start_cluster[i]))) {
				// find "P31" extension
				SD_findImage(SD.drive_start_cluster[i], (char *)0, "P31", 0, &(image.protect[i]), &(image.len[i]), &(image.start_cluster[i]));
			}
			// check lock switch
			if (bit_is_set(PINB,2)) image.protect[i] = 8;
			// set mode
			if (SD_findImage(SD.drive_start_cluster[i], "2D      ", (char *)0, 0, (char *)0, (unsigned long *)0, (unsigned short *)0)) {
				pc6031.mode |= pow2(i);
			}
		}
	}
	SD_cmd(16, (unsigned long)512);
}

// called when the card is inserted or removed
void SD_checkEject(void)
{
	unsigned long i;
	unsigned char j,k;
	unsigned char wt;

	if (bit_is_set(PINB,3)) {
		for (j=0; j<DRIVE_NUM; j++) image.start_cluster[j] = 0;
		SD.inited = 0;
	} else if (!SD.inited) {
		LED = 0b00000001;
		PORTB = ((PORTB&0b11111100) | LED);
		wt = (RST_init) ? 3 : 6;
		for (k=0; k!=wt;k++)
		for (i=0; i!=0x50000;i++)
			if (bit_is_set(PINB,3)) return;
		SD_init();
		SD.inited = 1;
		LED = 0b00000000;
		PORTB = ((PORTB&0b11111100) | LED);
	}
}

// communicate with PC-6001
// send data to PC-6001
void P6_snd(unsigned char d1, unsigned char d2)
{ 
	while (bit_is_clear(PIND,1)) ;	// wait RFD set
	PORTA = d1;
	PORTD |= 0b00010000;			// set DAV
	while (bit_is_clear(PIND,2)) ;	// wait DAC set
	PORTA = d2;
	PORTD &= 0b11101111;			// reset DAV
	while (bit_is_set(PIND,2)) ;	// wait DAC reset
}

// receive data from PC-6001
void P6_rcv(unsigned char *d1, unsigned char *d2)
{
	*d1 = 0xff;
	*d2 = 0xff;
	PORTD |= 0b00100000;								// set RFD
	while (bit_is_clear(PIND,0)) ;						// wait DAV set
	PORTD &= 0b11011111;								// reset RFD
	*d1 = PINC;
	PORTD |= 0b01000000;								// set DAC
	while (bit_is_set(PIND,0)) ;						// wait DAV reset
	if (d2) *d2 = PINC;
	PORTD &= 0b10111111;								// reset DAC
}

// PC interrupt is used for detecting SD card eject
ISR(PCINT1_vect)
{
	SD_checkEject();
}

// initialize PC-6031 status
void pc6031_init(void)
{
	pc6031.valid = 0;
	pc6031.error = 0;
	pc6031.mode = 0;
}

// process PC-6031 command
void pc6031_process(void)
{
	while (1) {
		unsigned char cmd;

		while (bit_is_clear(PIND,3)) ;	// wait attention
		P6_rcv(&cmd, 0);				// receive command
		while (bit_is_set(PIND,3)) ;	// wait not attention
		switch (cmd) {
		case 0:		// init
			pc6031.error = 0;	
			break;
		case 1:		// write
		case 17: {	// high speed write
			unsigned char lg, dr, trk, sec, i;
			unsigned short j;

			pc6031.error = 0;
			P6_rcv(&lg, 0); P6_rcv(&dr, 0); P6_rcv(&trk, 0); P6_rcv(&sec, 0);
			if (!image.start_cluster[dr]) { pc6031.error = 1; break; }
			LED = 0b00000010;
			PORTB = ((PORTB&0b11111100) | LED);	
			if (!(pow2(dr)&pc6031.mode)) trk<<=1;
			for (i=0; i!=lg; i++) {
				unsigned short long_sector;
				unsigned short ft;

				if (image.protect[dr]) {
					pc6031.error = 1;
				} else {
					if (!SD_prepareFat(&long_sector, &ft, trk, sec, dr)) {
						pc6031.error = 1;
					} else {
						PORTB |= (1<<SD_CS);
						PORTB &= ~(1<<SD_CS);
						SD_cmd(24, (unsigned long)SD.userAddr+((unsigned long)(ft-2)*SD.sectorsPerCluster
						+ (unsigned long)(long_sector%SD.sectorsPerCluster))*512);
						SD_writeByte(0xff);
						SD_writeByte(0xfe);
					}
				}
				if (cmd==1)
					for (j=0; j!=256; j++) {
						unsigned char d;
						
						P6_rcv(&d, 0);
						if (pc6031.error == 0) SD_writeByte(d);
					}
				else
					for (j=0; j!=128; j++) {
						unsigned char d1, d2;
						
						P6_rcv(&d1, &d2);
						if (pc6031.error == 0) {
							SD_writeByte(d1);
							SD_writeByte(d2);
						}
					}			
				if (pc6031.error == 0) {
					for (j=0; j!=256; j++) SD_writeByte(0);
					SD_writeByte(0xff);
					SD_writeByte(0xff);
					SD_readByte();
					SD_waitFinish();
					PORTB |= (1<<SD_CS);
					PORTB &= ~(1<<SD_CS);
				}
				if (sec == 16) sec=1; else sec++;
			}
			LED = 0;
			PORTB = ((PORTB&0b11111100) | LED);
			break;
		}
		case 2:		// read
			P6_rcv(&pc6031.length, 0);
			P6_rcv(&pc6031.drive, 0);
			P6_rcv(&pc6031.track, 0);
			P6_rcv(&pc6031.sector, 0);
			
			if (!image.start_cluster[pc6031.drive]) { pc6031.error=1; break; }
			LED = 0b00000001;
			PORTB = ((PORTB&0b11111100) | LED);
			if (!(pow2(pc6031.drive)&pc6031.mode)) pc6031.track<<=1;
			pc6031.valid = 1;
			pc6031.error = 0;
			LED = 0;
			PORTB = ((PORTB&0b11111100) | LED);
			break;
		case 3:		// transfer read data
		case 18: {	// high speed transfer read data
			unsigned char i;
			unsigned short j;

			pc6031.error = 0;
			if ((!pc6031.valid) || (!image.start_cluster[pc6031.drive])) { pc6031.error=1; break; }
			LED = 0b00000001;
			PORTB = ((PORTB&0b11111100) | LED);
			for (i=0; i!=pc6031.length; i++) {
				unsigned short long_sector;
				unsigned short ft;

				if (!SD_prepareFat(&long_sector, &ft, pc6031.track, pc6031.sector, pc6031.drive)) {
					pc6031.error = 1;
				}
				if (pc6031.error == 0) {
					SD_cmd17(SD.userAddr+((unsigned long)(ft-2)*SD.sectorsPerCluster
						+ (long_sector%SD.sectorsPerCluster))*512);
				}
				if (cmd==3) for (j=0; j!=256; j++) {
					unsigned char d;
					if (pc6031.error == 0) {
						d = SD_readByte();
					} else {
						d = 0xff;
					}
					
					P6_snd(d, 0);
				} else for (j=0; j!=128; j++) {
					unsigned char d1;
					unsigned char d2;
					if (pc6031.error == 0) {
						d1 = SD_readByte();
						d2 = SD_readByte();
					} else {
						d1 = 0xff;
						d2 = 0xff;
					}
					
					P6_snd(d1, d2);
				}
				if (pc6031.error == 0) {
					for (j=0; j!=256; j++) SD_readByte();
					SD_readByte(); SD_readByte(); // discard CRC
				}
				if (pc6031.sector == 16) pc6031.sector=1; else pc6031.sector++;
			}
			pc6031.valid = 0;
			LED = 0;
			PORTB = ((PORTB&0b11111100) | LED);
			break;
		}
		case 4:	{	// sector copy
			unsigned char len, i;
			unsigned char src_drv, src_trk, src_sec;
			unsigned char dst_drv, dst_trk, dst_sec;
			unsigned short j;
				
			pc6031.error = 0;
			P6_rcv(&len, 0);
			P6_rcv(&src_drv, 0); P6_rcv(&src_trk, 0); P6_rcv(&src_sec, 0);
			P6_rcv(&dst_drv, 0); P6_rcv(&dst_trk, 0); P6_rcv(&dst_sec, 0);

			if (!(pow2(src_drv)&pc6031.mode)) src_trk<<=1;
			if (!(pow2(dst_drv)&pc6031.mode)) dst_trk<<=1;
				
			if ((!image.start_cluster[src_drv])||!image.start_cluster[dst_drv]) {
				pc6031.error=1; break; }
				
			for (i=0; i<len; i++) {
				unsigned short long_sector;
				unsigned short ft;
				static unsigned char buff[256];

				LED = 0b00000001;
				PORTB = ((PORTB&0b11111100) | LED);

				if (!SD_prepareFat(&long_sector, &ft, src_trk, src_sec, src_drv)) {
					pc6031.error = 1;
					break;
				}
				SD_cmd17(SD.userAddr+((unsigned long)(ft-2)*SD.sectorsPerCluster
					+ (long_sector%SD.sectorsPerCluster))*512);
				for (j=0; j!=256; j++) buff[j] = SD_readByte();
				for (j=0; j!=256; j++) SD_readByte();
				SD_readByte(); SD_readByte(); // discard CRC
				
				if (src_sec == 16) src_sec=1; else src_sec++;

				LED = 0b00000010;
				PORTB = ((PORTB&0b11111100) | LED);

				if (!image.protect[dst_drv]) {
					if (!SD_prepareFat(&long_sector, &ft, dst_trk, dst_sec, dst_drv)) {
						pc6031.error = 1;
						break;
					}

					PORTB |= (1<<SD_CS);
					PORTB &= ~(1<<SD_CS);
					SD_cmd(24, (unsigned long)SD.userAddr+((unsigned long)(ft-2)*SD.sectorsPerCluster
						+ (unsigned long)(long_sector%SD.sectorsPerCluster))*512);
					SD_writeByte(0xff);
					SD_writeByte(0xfe);
					for (j=0; j!=256; j++) SD_writeByte(buff[j]);
					for (j=0; j!=256; j++) SD_writeByte(0);
					SD_writeByte(0xff);
					SD_writeByte(0xff);
					SD_readByte();
					SD_waitFinish();
					PORTB |= (1<<SD_CS);
					PORTB &= ~(1<<SD_CS);
				}
				if (dst_sec == 16) dst_sec=1; else dst_sec++;	
			}
			LED = 0;
			PORTB = ((PORTB&0b11111100) | LED);
			break;
		}
		case 5:	{	// format
			unsigned char trk, sec;
			unsigned char drv, is1d;
			unsigned short j;
				
			pc6031.error = 0;
			P6_rcv(&drv, 0);	
			if (image.protect[drv] || !image.start_cluster[drv]) { pc6031.error = 1; break; }
			is1d = (!(pow2(drv)&pc6031.mode));
			LED = 0b00000010;
			PORTB = ((PORTB&0b11111100) | LED);
			// initialize the area with 0xff
			for (trk=0; trk!=80; trk++) {
				if (is1d && (trk&1) ) continue;
				for (sec=1; sec!=17; sec++) {
					unsigned short long_sector;
					unsigned short ft;

					if (!SD_prepareFat(&long_sector, &ft, trk, sec, drv)) {
						pc6031.error = 1;
						break;
					}

					PORTB |= (1<<SD_CS);
					PORTB &= ~(1<<SD_CS);

					SD_cmd(24, (unsigned long)SD.userAddr+((unsigned long)(ft-2)*SD.sectorsPerCluster
						+ (unsigned long)(long_sector%SD.sectorsPerCluster))*512);
					SD_writeByte(0xff);
					SD_writeByte(0xfe);	

					// format with 0xff
					for (j=0; j!=256; j++) SD_writeByte(0xff);
					// dummy 256 bytes
					for (j=0; j!=256; j++) SD_writeByte(0);
					SD_writeByte(0xff);
					SD_writeByte(0xff);
					SD_readByte();
					SD_waitFinish();

					PORTB |= (1<<SD_CS);
					PORTB &= ~(1<<SD_CS);
				}
				if (pc6031.error == 1) {
					break;
				}
			}
			LED = 0;
			PORTB = ((PORTB&0b11111100) | LED);
			break;
		}
		case 6:		// command status
			P6_snd(0b10000000 | (pc6031.valid<<6) | pc6031.error, 0);
			pc6031.error = 0;
			break;
		case 7: {	// drive status
			const unsigned char status =
				(DRIVE_NUM==1)?0b00011111:((DRIVE_NUM==2)?0b00111111:
				((DRIVE_NUM==3)?0b01111111:0b11111111));

			P6_snd(status, 0);
			pc6031.error = 0;
			break;
		}
		case 11: {	// transfer FDD's memory data to PC-6001
					// only for address 7EFh (return EFh) & 75Fh (return 77h)
			unsigned short i;
			unsigned char AH, AL, DH, DL;

			P6_rcv(&AH, 0); P6_rcv(&AL, 0); P6_rcv(&DH, 0); P6_rcv(&DL, 0);
			P6_snd((AH==7)?((AL==0x5f)?0x77:((AL==0xef)?0xef:0xff)):0xff, 0);
			for (i = 1; i < ((((short)DH)<<8)|DL); i++) P6_snd(0xff, 0);
			pc6031.error = 0;
			break;
		}
		case 20: {
			unsigned char drv;
			
			P6_rcv(&drv, 0);
			P6_snd(image.protect[drv]?0b01000000:0,0);
			pc6031.error = 0;
			break;
		}
		case 21: {
			pc6031.error = 0;
			break;
		}
		case 23: {	// set mode (1D or 2D)
			unsigned char d;
			P6_rcv(&d, 0);
			pc6031.mode =(d&0b00001111);
			pc6031.error = 0;
			break;
		}
		case 24: {	// ask mode
			P6_snd(pc6031.mode, 0);
			pc6031.error = 0;
			break;
		}
		case 0xFD: { // send P31 image names to P6
			unsigned char drv;

			P6_rcv(&drv, 0);
			LED = 0b00000001;
			PORTB = ((PORTB&0b11111100) | LED);
			sendFileNames(SD.drive_start_cluster[drv]);
			SD_cmd(16, (unsigned long)512);
			pc6031.error = 0;
			LED = 0;
			PORTB = ((PORTB&0b11111100) | LED);
			break;
		}
		case 0xFE: { // change image
			unsigned char drv, i, res;
			char name[9];

			pc6031.error = 1;
			P6_rcv(&drv, 0);
			LED = 0b00000001;
			PORTB = ((PORTB&0b11111100) | LED);
			for (i=0; i!=8; i++)
				P6_rcv((unsigned char *)&(name[i]), 0);
			name[8]=0;
			res = SD_findImage(SD.drive_start_cluster[drv], name, "P31", 0, &(image.protect[drv]), &(image.len[drv]), &(image.start_cluster[drv]));
			if (res) {
				image.prevFatNum[drv] = 0xff;
				pc6031.error = 0;
				if (SD_findImage(SD.drive_start_cluster[drv], "2D      ", (char *)0, 0, (char *)0, (unsigned long *)0, (unsigned short *)0)) {
					pc6031.mode |= pow2(drv);
				}
			}
			SD_cmd(16, (unsigned long)512);
			LED = 0;
			PORTB = ((PORTB&0b11111100) | LED);
			break;
		}		
		}
	}
}

int main(void)
{	
	DDRB = 0b10110011;  // CLK,DO,DI,CS,EJ,WP,LED2,LED1
	DDRA = 0b11111111;	// data out to P6
	DDRC = 0b00000000;	// data in from P6
	DDRD = 0b11110000;	// ctrl out to P6 (ATN,DAC,RFD,DAV),
						// ctrl in from P6 (ATN,DAC,RFD,DAV)

	PORTB = 0b01001100;	// pull up setting for inputs (DO, EJ, WP)
	PORTA = 0b00000000;
	PORTC = 0b00000000;
	PORTD = 0b00000000;	// ATN output is always 0

	// PC int is used for SD card's eject
	PCMSK1 |= 0x08;
	PCICR |= (1<<PCIE1);

	RST_init = 1;
	LED = 0;
	PORTB = ((PORTB&0b11111100) | LED);

	pc6031_init();
	
	SD.inited = 0;
	SD_checkEject();
	
	RST_init = 0;

	sei();
	
	pc6031_process();
}
